@@ -10,7 +10,7 @@
 # 4) Overlap file extents
 # 5) Unable to rollback
 
-source "$TOP/tests/common"
+source "$TEST_TOP/common"
 
 check_prereq btrfs-convert
 check_prereq btrfs
@@ -18,10 +18,10 @@ check_global_prereq e2fsck
 check_global_prereq xzcat
 
 setup_root_helper
-prepare_test_dev 512M
+prepare_test_dev
 
 # override common function
-function check_image() {
+check_image() {
 	TEST_DEV="$1"
 	run_check e2fsck -n -f "$TEST_DEV"
 	run_check "$TOP/btrfs-convert" "$TEST_DEV"
